#ifndef FO_PARAMETER_H
#define FO_PARAMETER_H

#include "Kinematics.h"


namespace CuTe {
namespace NS_FO{


/* \class FO_NLO_Kinematic_Parameter
 * \brief NLO Kinematic_Parameter
 *
 * Independant of scale mu, thus calculate only once for all cross-sections/pdf-sets
 *
 */
using NS_Utils::NS_pow::pow_2;
using NS_Kinematics::STUX2;
using NS_Kinematics::Kinematics;

using std::cerr;
using std::endl;
using std::swap;


class FO_NLO_Kinematic_Parameter {
public:
	//init
	double S;
	double sS;	//=sqrt(S)

	double M2;
	double iM2;	//= 1./M2

	//set_qT_mu
	double qT2;
	double iqT2;
	double mT2;	//= M2+qT2
	double mT;	//= sqrt(mT2)
	double lT2;// (qT/mT)^2

	//set_y
	double T;
	double U;
	double X2;
	double iX2,fX2;	//= S+T+U-M2	= log(X2/M2)
	double _xip,_xim;	//= (U-M2)/(-S) = (T-M2)/(-S)

	//update
	double xi[2];
	double s,t,u,x2;

	//_update
	double A,B,C,iA;
	//A=U + x0(S2-U)	= max[x2](xi_0)//(2.7)
	//B= -U/(X2-U)	= min[xi_0]//(2.7)
	//C=-t/(S xi[0] - (M2-U) )	= min[xi_1](xi_0)
	double s2,is,is2;	//= s^2, 1/s, 1/s^2
	double t2,t4,it,it2;
	double u2,u4,iu,iu2;
	double x4,ix2;
	double x2T,ix2T;	//=x2+qT2
	double Jac_1d;			//= D[xi0,x0] * 1/xi0	* { D[xi1,x2]	*1/xi1 }@{x2=0}
	double Jac_2d;			//= D[xi0,x0] * 1/xi0	* D[xi1,x1]		*1/xi1


	inline virtual void init(const Kinematics *Kin){
//		NS_Utils::DOUT dout("FO_NLO_Kinematic_Parameter::init","Kin");
		S	= Kin->S;
		sS	= Kin->sS;
		M2	= Kin->M2;

		iM2		= 1./M2;
	}

	inline void set_qT(const Kinematics *Kin){
//		NS_Utils::DOUT dout("FO_NLO_Kinematic_Parameter::set_qT","Kin");
		qT2	= Kin->qT2;
		iqT2= 1./qT2;
		mT2	= M2+qT2;
		mT	= sqrt(mT2);
		lT2	= qT2/mT2;
	}

	inline void set_y(const Kinematics *Kin){
//		NS_Utils::DOUT dout("FO_NLO_Kinematic_Parameter::set_y","Kin");

		const STUX2 *stux2	= &(Kin->stux2);

		T	= stux2->T;
		U	= stux2->U;
		X2	= stux2->X2;

//		if((S + U + T - M2 - X2)!=0.)	cerr	<<"H_NLO::Parameter::set_y X2 Delta="<<(S + U + T - M2 - X2)<<"\t Ratio="<<(S + U + T - M2 - X2)/X2<<endl;// DEBUG

		iX2	= 1./X2;
		fX2	= log(X2*iM2);
		_xip= stux2->xi_pm[0];
		_xim= stux2->xi_pm[1];

	}

protected:

	inline virtual void _update_nnlo(){
		cerr<< "********************************************************************" << endl
			<< "*** FO_NLO_Kinematic_Parameter::_update_nnlo() called -> exit(1) ***" << endl
			<< "********************************************************************" << endl;
		exit(1);
	}

	template < int order >
	inline void _update(const STUX2 *stux2){
//		NS_Utils::DOUT dout("FO_NLO_Kinematic_Parameter::_update","stux2");

		B	= _xip*( 1. + lT2 * _xip/(1.-_xim) );
		C	= _xim*( 1. + lT2 * _xip/(xi[0]-_xip) );
		A	= S*(xi[0]-_xip)*(1.-C);
		iA	= 1./A;

		s2		= s*s;
		is		= 1./s;
		is2		= is*is;

		t2		= t*t;
		t4		= t2*t2;
		it		= 1./t;
		it2		= it*it;

		u2		= u*u;
		u4		= u2*u2;
		iu		= 1./u;
		iu2		= iu*iu;

		x4		= x2*x2;
		ix2		= ( (x2!=0.) ? (1./x2) : (0.) );

		x2T		= qT2 + x2;
		ix2T	= 1./x2T;

		Jac_2d	= stux2->Jac;
		Jac_1d	= stux2->Jac0;

		if(order>1) _update_nnlo();
	}

public:
	template < int order >
	inline void update_0(const Kinematics *Kin){
//		NS_Utils::DOUT dout("FO_NLO_Kinematic_Parameter::update_0","Kin");
		const STUX2 *stux2	= &(Kin->stux2);
		xi[0]	= stux2->xi0[0];
		xi[1]	= stux2->xi0[1];

		s	= stux2->s0;
		t	= stux2->t0;
		u	= stux2->u0;
		x2	= stux2->x20;

		_update<order>(stux2);
	}

	template < int order >
	inline void update(const Kinematics *Kin){
//		NS_Utils::DOUT dout("FO_NLO_Kinematic_Parameter::update","Kin");
		const STUX2 *stux2	= &(Kin->stux2);

		xi[0]	= stux2->xi[0];
		xi[1]	= stux2->xi[1];

		s	= stux2->s;
		t	= stux2->t;
		u	= stux2->u;
		x2	= stux2->x2;

		_update<order>(stux2);
	}


	inline virtual void swap_t_u(){
//		NS_Utils::DOUT dout("FO_NLO_Kinematic_Parameter::swap_t_u");
		swap(t,u);
		swap(t2,u2);
		swap(t4,u4);
		swap(it,iu);
		swap(it2,iu2);
	}

	virtual std::ostream& properties(std::ostream &os, const int ntabs)const;

	virtual ~FO_NLO_Kinematic_Parameter(){

	}

};



/* \class FO_NNLO_Kinematic_Parameter
 * \brief NNLO FO_NNLO_Kinematic_Parameter
 *
 * Independant of scale mu, thus calculate only once for all cross-sections/pdf-sets
 *
 */
class FO_NNLO_Kinematic_Parameter: public FO_NLO_Kinematic_Parameter {
public:
//	const Plus_x2 *plus_x2_A,*plus_x2_X2;

	double M2t,iM2t;	//= M2-t
	double M2u,iM2u;	//= M2-u

	double ids,idt,idu;			//=s+M2-x2==AGS//=x2-t//=x2-u
	double ds,dt,du;	//=1./(s+M2-x2)//=1./(x2-t)//=1./(x2-u)

	double za;	//=-t/(x2-t)=-t*dt
	double iza,iza2,iza4;
	double zb;	//=-u/(x2-u)=-u*du
	double izb,izb2,izb4;

	double GS_B2;		//= ids^2 - 4 M2 s
	double GS_B;
	double GS_iB;
	double GS_iB2;
	double GS_iB4;
	double dst,dsu,dtu;	//=1./(s-idt);=1./(s-idu);=1./(t*u-x2*M2)=1/(qT2*s);
	double idtu2;		//=(t*u-x2*M2)^2
	double l,il,il2,il3,il4,il5;
	double fs,ft,fu;	//=log( s*iM2);log(-t*iM2);log(-u*iM2);
	double fx2,fA;	//=log( x2*iM2);log(mu2*iM2);log(A*iM2);
	double fst,fsu,ftu,fstu;	//=log(s*t*t*iM2*dt2);log(s*u*u*iM2*du2);log(idtu*dt*du);log(s*M2*dt*du)=-fx2p;
	double fl;
	double flt,flu;
	double f1t,f2t,f1u,f2u;	//f1x = Log_2(M2/(M2-x)) + 1/2 * Log^2(M2/(M2-x))

	double fL2a;	//	= log(M2*s/((ids+s*t*dt)*(ids+s*t*dt)));
	double fL2b;	//	= log(M2*s/((ids+s*u*du)*(ids+s*u*du)));
	double fL3;		//	= log((ids+GS_B)/(ids-GS_B));

	double LtX2,LuX2;	//= log(-t/X2)	//= log(-u/X2)
	double Lx2pX2;		//= log((x2+qT2)/X2)



	inline virtual void init(const Kinematics *Kin){
//		NS_Utils::DOUT dout("FO_NNLO_Kinematic_Parameter::init","Kin");
		FO_NLO_Kinematic_Parameter::init(Kin);
//		plus_x2_A	= &(Kin->stux2.plus_x2_A);
//		plus_x2_X2	= &(Kin->stux2.plus_x2_X2);
	}


	inline void swap_t_u(){
//		NS_Utils::DOUT dout("FO_NNLO_Kinematic_Parameter::swap_t_u");
		FO_NLO_Kinematic_Parameter::swap_t_u();

		swap(M2t,M2u);
		swap(iM2t,iM2u);

		swap(dt,du);
		swap(idt,idu);
		swap(dst,dsu);

		swap(ft,fu);
		swap(fst,fsu);

		swap(flt,flu);
		swap(f1t,f1u);
		swap(f2t,f2u);

		swap(za,zb);
		swap(iza,izb);
		swap(iza2,izb2);
		swap(iza4,izb4);

		swap(fL2a,fL2b);

		swap(LtX2,LuX2);
	}

	void _update_nnlo();


	virtual std::ostream& properties(std::ostream &os, const int ntabs)const;
};

class FO_Scale_Parameter {
public:
	const double *_mu;
	const double *_as;
	const int	 *_nf;

	double muF;
	double muR;
	double as;
	int	   nf;

//	double fmu;
	double fmuF; 	//= log(muF2*iM2);
	double fmuR; 	//= log(muR2*iM2);
	double LmuX2;	//= log(mu2*iX2);
	double Ltumu;	//= log((t*u)/pow_2(mu2));

	inline void init(const double *_mu, const double *_as, const int *_nf){
		this->_mu	= _mu;
		this->_as	= _as;
		this->_nf	= _nf;

		muF	= *_mu;
		muR	= *_mu;
		as	= *_as;
		nf	= *_nf;

//		fmu		= nan;
//		fmuF	= nan;
//		fmuR	= nan;
//		LmuX2	= nan;
//		Ltumu	= nan;
	}

	void update(const FO_NNLO_Kinematic_Parameter* p){
		muF	= *_mu;
		muR	= *_mu;
		as	= *_as;
		nf	= *_nf;

		const double mu2	= pow_2(*_mu);
		const double fmu	= log(mu2*p->iM2);
		fmuF	= fmu;
		fmuR	= fmu;
		LmuX2	= log(mu2*p->iX2);
		Ltumu	= log((p->t)*(p->u)/pow_2(mu2));
	}
};




}  // namespace NS_FO
}  // namespace CuTe

#endif

