#include <iostream>
#include <vector>
#include <cmath>

#include "Stepfunction.h"
#include "param_options.h"

#include <boost/math/special_functions/erf.hpp>

namespace CuTe {
namespace NS_Stepfunction {

using std::cerr;
using std::cout;
using std::endl;
using std::ostream;
using std::string;
using std::swap;
using std::vector;

ostream& Sigmoid::to_stream(ostream &os, const int ntabs)const{
	string tab;
	for(int i=0; i<ntabs; ++i) tab += "\t";
	os<<tab	<<"\t"<<"m = "<<m
			<<"\t"<<"w = "<<w
			<<endl;
	return os;
}

class Sigmoid_No: public Sigmoid{
public:
	inline double f(const double x)const{
		return 1.;
	}

	inline double barf(const double x)const{
		return 1.;
	}

	inline Sigmoid_No()
	:Sigmoid(){
	}

	inline Sigmoid_No(const Sigmoid_No& S)
	:Sigmoid(S){
	}

	inline Sigmoid_No& operator=(const Sigmoid_No& S){
		Sigmoid::operator =(S);
		return *this;
	}

	inline ~Sigmoid_No(){
	}

	ostream& to_stream(ostream &os, const int ntabs)const{
		string tab("");
		for(int i=0; i<ntabs; ++i)	tab+="\t";

		os	<<tab<<"\tf(x) = 1."<<endl;
		return os;
	}

};

class Sigmoid_Logistic: public Sigmoid{
protected:
	double C;

	Sigmoid_Logistic();

public:
	inline double f(const double x)const{
		return 1./(1.+exp(+C*(x-m)));
	}

	inline double barf(const double x)const{
		return 1./(1.+exp(-C*(x-m)));
	}

	Sigmoid_Logistic(const double l, const double r, const double R)
	:Sigmoid(l,r,R)
	,C(log((1.-R)/R)/w){
	}

	ostream& to_stream(ostream &os, const int ntabs)const{
		string tab("");
		for(int i=0; i<ntabs; ++i)	tab+="\t";

		os	<<tab<<"\tf(x) = 1./(1.+exp(C*(x-m)))"<<endl;
		os	<<tab<<"\tC = log((1.-r)/r)/w = "<<C<<endl;
		Sigmoid::to_stream(os,ntabs);
		return os;
	}

};

class Sigmoid_Erf: public Sigmoid{
protected:
	double C;

	Sigmoid_Erf();

public:
	inline double f(const double x)const{
		return 0.5*(1.-erf(C*(x-m)));
	}

	inline double barf(const double x)const{
		return 0.5*(erf(C*(x-m))-1.);
	}

	Sigmoid_Erf(const double l, const double r, const double R)
	:Sigmoid(l,r,R)
	,C(boost::math::erf_inv(1.-2.*R)/w){
	}

	ostream& to_stream(ostream &os, const int ntabs)const{
		string tab("");
		for(int i=0; i<ntabs; ++i)	tab+="\t";

		os	<<tab<<"\tf(x) = 0.5*(1.-erf(C*(x-m)))"<<endl;
		os	<<tab<<"\tC = erf^{-1}(1.-2.*R)/w) = "<<C<<endl;
		Sigmoid::to_stream(os,ntabs);
		return os;
	}

};

class Sigmoid_Gompertz: public Sigmoid{
protected:
	double A;
	double B;

	Sigmoid_Gompertz();

public:
	inline double f(const double x)const{
		return exp(-A*exp(-B*(x-m)));
	}

	inline double barf(const double x)const{
		return 1.-f(x);
	}

	Sigmoid_Gompertz(const double l, const double r, const double R)
	:Sigmoid(l,r,R)
	,A(sqrt(log(1.-R)*log(R)))
	,B(log(-A/log(R))/w){
	}

	ostream& to_stream(ostream &os, const int ntabs)const{
		string tab("");
		for(int i=0; i<ntabs; ++i)	tab+="\t";

		os	<<tab<<"\tf(x) = 1.-exp(-A*exp(-B*(x-m)))"<<endl;
		os	<<tab<<"\tA = sqrt(log(1.-R)*log(R)) = "<<A<<"\tB = log(-A/log(R))/w = "<<B<<endl;
		Sigmoid::to_stream(os,ntabs);
		return os;
	}

};

class Sigmoid_Gompertz_Mirrored: public Sigmoid{
protected:
	double A;
	double B;

	Sigmoid_Gompertz_Mirrored();

public:
	inline double f(const double x)const{
		return 1.- barf(x);
	}

	inline double barf(const double x)const{
		return exp(-A*exp(+B*(x-m)));
	}

	Sigmoid_Gompertz_Mirrored(const double l, const double r, const double R)
	:Sigmoid(l,r,R)
	,A(sqrt(log(1.-R)*log(R)))
	,B(log(-A/log(R))/w){
	}

	ostream& to_stream(ostream &os, const int ntabs)const{
		string tab("");
		for(int i=0; i<ntabs; ++i)	tab+="\t";

		os	<<tab<<"\tf(x) = 1.-exp(-A*exp(+B*(x-m)))"<<endl;
		os	<<tab<<"\tA = sqrt(log(1.-R)*log(R)) = "<<A<<"\tB = log(-A/log(R))/w = "<<B<<endl;
		Sigmoid::to_stream(os,ntabs);
		return os;
	}

};


void TransitionFunction::_exit(){
	if(S!=NULL){
		switch (type) {
			case TFT_LOG:
				delete static_cast<Sigmoid_Logistic*>(S);
				break;
			case TFT_ERF:
				delete static_cast<Sigmoid_Erf*>(S);
				break;
			case TFT_GPZ:
				delete static_cast<Sigmoid_Gompertz*>(S);
				break;
			case TFT_MGPZ:
				delete static_cast<Sigmoid_Gompertz_Mirrored*>(S);
				break;
			case TFT_NO:
				delete static_cast<Sigmoid_No*>(S);
				break;
			case TFT_WRONG:
			default:
				cerr	<<"TransitionFunction::_exit() WRONG TF to delete"<<endl;
				break;
		}
		S	= NULL;
	}
}


bool TransitionFunction::set_to_one(){
	_exit();

	S	= new Sigmoid_No();
	iS0	= 1.;
	type= TFT_NO;

	left		= 0;
	right		= 0;
	Reduction	= 0;
//	double M,CB;
	return true;
}

void TransitionFunction::_init(const ENUM_Transition_Function_Type type){
	_exit();
	this->type	= type;
	iS0	= 1.;
	switch (type) {
		case TFT_LOG:
			S	= new Sigmoid_Logistic(left,right,Reduction);
			iS0	= 1./f(0);
			break;
		case TFT_ERF:
			S	= new Sigmoid_Erf(left,right,Reduction);
			iS0	= 1./f(0);
			break;
		case TFT_GPZ:
			S	= new Sigmoid_Gompertz(left,right,Reduction);
			iS0	= 1./f(0);
			break;
		case TFT_MGPZ:
			S	= new Sigmoid_Gompertz_Mirrored(left,right,Reduction);
			iS0	= 1./f(0);
			break;
		case TFT_NO:
			S	= new Sigmoid_No();
			break;
		case TFT_WRONG:
		default:
			cerr	<<"TransitionFunction::_init() WRONG TF to create => NO"<<endl;
			S	= new Sigmoid_No();
			break;
	}
}

bool TransitionFunction::init(vector<double>::const_iterator lrR, const double M, const double CB, ENUM_Transition_Function_Type type, const ENUM_Transition_Function_Arg arg){
	this->arg	= arg;
	this->M		= M;
	this->CB	= CB;

	if(type==TFT_NO) return set_to_one();

	this->left		= *(lrR++);
	this->right		= *(lrR++);
	this->Reduction	= *(lrR++);

	if( (left==right)||(Reduction<=0)||(1<=Reduction) ){
		cerr	<<"TransitionFunction::init() TF args Wrong"<<endl
				<<"\t"<<"left = "<<left<<endl
				<<"\t"<<"right = "<<right<<endl
				<<"\t"<<"Reduction = "<<Reduction<<endl
				<<"\t"<<"=> NO"<<endl;
		type	= TFT_WRONG;
	}

	switch (arg) {
		case TFA_ETA:
			if(left < right) swap(left,right);
			break;
		case TFA_LAMBDA:
		case TFA_LAMBDA2:
			if(left > right) swap(left,right);
			if(left <= 0.){
				cerr	<<"TransitionFunction::init() TF args Wrong"<<endl
						<<"\t"<<"left = "<<left<<endl
						<<"\t"<<"right = "<<right<<endl
						<<"\t"<<"Reduction = "<<Reduction<<endl
						<<"\t"<<"=> NO"<<endl;
				type	= TFT_WRONG;
				return !set_to_one();
			}
			break;
		default:
			cerr	<<"TransitionFunction::init() WRONG TF -> NO"<<endl;
			return !set_to_one();
			break;
	}

	_init(type);
	return true;
}

bool TransitionFunction::init(const SOPS_CS_Params *OP,const ENUM_Transition_Function_Type type, const ENUM_Transition_Function_Arg arg){
	switch (OP->Process.prod()) {
		case NS_enums::NS_PROD::PROD_Higgs:
		case NS_enums::NS_PROD::PROD_H_RS:
			CB	= 3.;
			break;
		default:
			CB	= 4./3.;
			break;
	}
	return init(OP->Scheme.transition_to_FO().begin(),OP->Process.M(),CB,type,arg);
}

bool TransitionFunction::init(const SOPS_CS_Params *OP){
	return init(OP,OP->Scheme.transition_type(),OP->Scheme.transition_arg());
}


double TransitionFunction::f(const double qT)const{
	const double lambda(qT/M);
	switch (arg) {
		case TFA_LAMBDA:
			return S->f(lambda)*iS0;
			break;
		case TFA_LAMBDA2:
			return S->f(lambda*lambda)*iS0;
			break;
		case TFA_ETA:
			return S->f(eta_of(lambda))*iS0;
			break;
		default:
			return 0.;
			break;
	}
}


ostream& TransitionFunction::to_stream(ostream &os, const int ntabs)const{
	string tab;
	for(int i=0; i<ntabs; ++i) tab += "\t";


	os	<<tab<<"TransitionFunction ("<<type<<"("<<arg<<"), "<<left<<", "<<right<<", "<<Reduction<<"):"<<endl;
	S->to_stream(os,ntabs);
	if(type!=TFT_NO){
		switch (arg) {
		case TFA_ETA:
			os<<tab	<<"\t"<<"      "
					<<"\t"<<"qT[GeV]"
					<<"\t"<<"eta[alpha_s]"
					<<"\t"<<"lambda"
					<<endl;
			os<<tab	<<"\t"<<"left :"
					<<"\t"<<M*lambda_of(left)
					<<"\t"<<left
					<<"\t"<<lambda_of(left)
					<<endl;
			os<<tab	<<"\t"<<"right:"
					<<"\t"<<M*lambda_of(right)
					<<"\t"<<right
					<<"\t"<<lambda_of(right)
					<<endl;
			os<<tab	<<endl;
			os<<tab	<<"\t"<<"f(left ) = "<<100.*f(M*lambda_of(left))<<"%;"
					<<"\t"<<"f(right) = "<<100.*f(M*lambda_of(right))<<"%;"
					<<endl;
			break;
		case TFA_LAMBDA:
			os<<tab	<<"\t"<<"      "
					<<"\t"<<"qT[GeV]"
					<<"\t"<<"lambda"
					<<"\t"<<"eta[alpha_s]"
					<<endl;
			os<<tab	<<"\t"<<"left :"
					<<"\t"<<M*(left)
					<<"\t"<<left
					<<"\t"<<eta_of(left)
					<<endl;
			os<<tab	<<"\t"<<"right:"
					<<"\t"<<M*(right)
					<<"\t"<<right
					<<"\t"<<eta_of(right)
					<<endl;
			os<<tab	<<endl;
			os<<tab	<<"\t"<<"f(left ) = "<<100.*f(M*(left))<<"%;"
					<<"\t"<<"f(right) = "<<100.*f(M*(right))<<"%;"
					<<endl;
			break;
		case TFA_LAMBDA2:
			os<<tab	<<"\t"<<"      "
					<<"\t"<<"qT[GeV]"
					<<"\t"<<"lambda"
					<<"\t"<<"eta[alpha_s]"
					<<endl;
			os<<tab	<<"\t"<<"left :"
					<<"\t"<<M*sqrt(left)
					<<"\t"<<sqrt(left)
					<<"\t"<<eta_of(sqrt(left))
					<<endl;
			os<<tab	<<"\t"<<"right:"
					<<"\t"<<M*sqrt(right)
					<<"\t"<<sqrt(right)
					<<"\t"<<eta_of(sqrt(right))
					<<endl;
			os<<tab	<<endl;
			os<<tab	<<"\t"<<"f(left ) = "<<100.*f(M*sqrt(left))<<"%"
					<<"\t"<<"f(right) = "<<100.*f(M*sqrt(right))<<"%"
					<<endl;
			break;
		default:
			break;
		}
		os<<tab	<<"\t"<<"M  = "<<M
				<<"\t"<<"CB = "<<CB
				<<"\t"<<"S0 = "<<100./iS0<<"%"
				<<endl;
	}

	return os;
}


}	// namespace NS_Stepfunction
}  // namespace CuTe
