/*
 * lhafunc.cpp
 *
 *  Created on: May 1, 2014
 *      Author: daniel
 */

#include "lhafunc.h"
#include <LHAPDF/LHAPDF.h>

#ifdef __cplusplus
extern "C" {
#endif
	#include "gsl/gsl_roots.h"
	#include "gsl/gsl_errno.h"
#ifdef __cplusplus
}
#endif

using std::complex;
using CuTe::NS_LHAFUNC::A_strong;
using CuTe::NS_LHAFUNC::PDF;
using CuTe::NS_LHAFUNC::LHA_PDFSET;
using CuTe::NS_LHAFUNC::PDFUncertainty;
using CuTe::NS_Utils::GSL_ERROR_CODE;

void CuTe::NS_LHAFUNC::OWN_LHAPDF::del(){
	own	= false;
	delete pdf;		pdf	= NULL;
	delete alphas;	alphas	= NULL;
}


std::ostream& CuTe::NS_LHAFUNC::operator<<(std::ostream& os, const OWN_LHAPDF& my_lhapdf){
	os<<"my_lhapdf("<<my_lhapdf.pdf<<","<<my_lhapdf.alphas<<") = "<<my_lhapdf();
	return os;
}

const double CuTe::NS_LHAFUNC::A_strong::gen	= 1e-12;



void A_strong::init(const LHAPDF::AlphaS& lha_alpha_s, const int Order_QCD, const double estimate){
//		NS_Utils::DOUT dout("A_strong::init");
	this->lha_alpha_s	= &lha_alpha_s;

	p.Order_QCD		= Order_QCD;
	p.mu_0			= 0.;
	p.a_0			= 0.;
	p.nf			= 0;

	p.Lambda_QCD	= estimate;
}


double A_strong::as(const double mu)const{ return 0.25*M_1_PI*lha_alpha_s->alphasQ(mu);}

int A_strong::nf(double mu)const{
//		NS_Utils::DOUT dout("A_strong::nf","mu");
//		dout()	<<"this="<<this	<<endl;
//		dout()	<<"lha_alpha_s="<<lha_alpha_s	<<endl;
//		dout()	<<"mu="<<mu	<<endl;
//		dout()	<<"nf(mu/2)="<<lha_alpha_s->numFlavorsQ(mu/2.)	<<endl;
	int nf(lha_alpha_s->numFlavorsQ(mu));
//		dout()	<<"nf="<<nf	<<endl;
//	if((nf<1)||(nf>6)){	// debug
//		std::cerr	<<"nf("<<mu<<") = "<<nf	<<endl;
//		exit(1);
//	}
	return nf;
}


complex<double> A_strong::as(const complex<double> &mu)const{
//	NS_Utils::DOUT dout("A_strong::as","C(mu)");
	if(mu.imag()==0.) return as(mu.real());
	else{
//		_A_strong_parameter p;
		get_Lambda_QCD(abs(mu));
//		dout()	<<"mu="<<mu	<<"\t"<<"a0="<<p.a_0	<<"\t"<<"ord="<<p.order<<"="<<OrderQCD	<<"\t"<<"nf="<<p.Beta.get_n_f()	<<"\t"<<"Lambda="<<p.Lambda_QCD	<<endl;

		const complex<double> L = (log(mu/p.Lambda_QCD)*2.);
		return _a< complex< double > >(L,p.Beta,p.Order_QCD);
	}
}

double A_strong::quarkMass(const Parton& q)const{
	double quarkMass;
	int id(q.flv());
	try {
		quarkMass	= lha_alpha_s->quarkMass(id);
	} catch (const LHAPDF::Exception& e) {
		quarkMass	= lha_alpha_s->quarkThreshold(id);
	}
	return quarkMass;
}

double A_strong::quarkThreshold(const Parton& q)const{
	double quarkThreshold;
	int id(q.flv());
	try {
		quarkThreshold	= lha_alpha_s->quarkThreshold(id);
	} catch (const LHAPDF::Exception& e) {
		quarkThreshold	= lha_alpha_s->quarkMass(id);
	}
	return quarkThreshold;
}

double A_strong::get_Lambda_QCD(const double mu)const{
	const int new_nf(nf(mu));

//	if(new_nf!=p.nf){
		p.nf	= new_nf;
		p.a_0	= as(mu);
		p.mu_0	= mu;

		p.Beta.select(new_nf);

		_calc_Lambda();
//	}

	return p.Lambda_QCD;
}


int A_strong::_calc_Lambda()const{
	const int max_iter(1000);
	int status;

	double x0;
	double x = 2.*log(p.mu_0/p.Lambda_QCD);
	const double acc_abs(0.);
	const double acc_rel(gen);

	gsl_function_fdf FDF;
//	FDF.f		= &f;
//	FDF.df		= &df;
//	FDF.fdf		= &fdf;
	FDF.f		= &fL;
	FDF.df		= &dfL;
	FDF.fdf		= &fdfL;
	FDF.params	= &p;

	const gsl_root_fdfsolver_type* T(gsl_root_fdfsolver_steffenson);
//	const gsl_root_fdfsolver_type* T(gsl_root_fdfsolver_newton);
	gsl_root_fdfsolver* s(gsl_root_fdfsolver_alloc (T));
	gsl_root_fdfsolver_set(s, &FDF, x);

	int iter = 0;
	do{
		iter++;
		status = gsl_root_fdfsolver_iterate (s);
//			GSL_ERROR_CODE(status);
		x0 = x;
		x = gsl_root_fdfsolver_root (s);
		status = gsl_root_test_delta (x, x0, acc_abs, acc_rel);

		if(status!=GSL_SUCCESS && status!=GSL_CONTINUE){
//				cout << "A_strong::init_lambda()\t";
			GSL_ERROR_CODE("A_strong::init_lambda()",status);
		}
//			GSL_ERROR_CODE(status);
	}while (status == GSL_CONTINUE && iter < max_iter);
	gsl_root_fdfsolver_free (s);
	p.Lambda_QCD = p.mu_0*exp(-0.5*x);
	if(status!=GSL_SUCCESS){
		GSL_ERROR_CODE("A_strong::init_lambda()",status);
		std::cerr	<<"A_strong::_calc_Lambda() status="<<status	<<"\titer="<<iter<<"("<<max_iter<<")"<<endl
					<<"\t"<<"Lambda_QCD(mu="<<p.mu_0<<",nf="<<p.nf<<") = "<<x<<" + "<<p.mu_0*exp(-0.5*(acc_abs+acc_rel*x))<<" - "<<p.mu_0*exp(+0.5*(acc_abs+acc_rel*x))<<endl
					<<"\t"<<"as(L_QCD)-as_0 = "<<f(p.Lambda_QCD,&p)<<endl
					<<"\t"<<"as(L_QCD)="<<_a<double>(log(gsl_pow_2(p.mu_0/p.Lambda_QCD)),p.Beta,p.Order_QCD)<<"\tas_0="<<p.a_0<<endl;
	}
//		cout << "\t\t<- A_strong::init_lambda("<<estimate<<")  Lambda_QCD = " << Lambda_QCD << endl;
	return status;
}



//std::ostream& CuTe::NS_LHAFUNC::operator<<(std::ostream& os, const CuTe::NS_LHAFUNC::Parton& q){
//	switch (q.lha()) {
//		case -6:
//			os << "bar(t)";
//			break;
//		case -5:
//			os << "bar(b)";
//			break;
//		case -4:
//			os << "bar(c)";
//			break;
//		case -3:
//			os << "bar(s)";
//			break;
//		case -2:
//			os << "bar(u)";
//			break;
//		case -1:
//			os << "bar(d)";
//			break;
//		case 0:
//			os << "G";
//			break;
//		case 1:
//			os << "d";
//			break;
//		case 2:
//			os << "u";
//			break;
//		case 3:
//			os << "s";
//			break;
//		case 4:
//			os << "c";
//			break;
//		case 5:
//			os << "b";
//			break;
//		case 6:
//			os << "t";
//			break;
//		default:
//			os << "!!WRONG FLAVOUR!!";
//			break;
//	}
//	return os;
//}
//
//
//std::ostream& CuTe::NS_LHAFUNC::operator<<( std::ostream& os,const CuTe::NS_LHAFUNC::CKM_MATRIX& ckm ){
////	std::stringstream ss;
//	os	<<*(ckm.CVud)<<" "
//		<<*(ckm.CVus)<<" "
//		<<*(ckm.CVub)<<" "
//		<<*(ckm.CVcd)<<" "
//		<<*(ckm.CVcs)<<" "
//		<<*(ckm.CVcb)<<" "
//		<<*(ckm.CVtd)<<" "
//		<<*(ckm.CVts)<<" "
//		<<*(ckm.CVtb)<<" "
//		;
////	if(os!=cout) cout	<<"operator<<:"<<ss.str()	<<endl;
////	os << ss.str();
//	return os;
//}
//
//
//std::istream& CuTe::NS_LHAFUNC::operator>>( std::istream& is,CuTe::NS_LHAFUNC::CKM_MATRIX& ckm ){
//	is	>>*(ckm.Vud)
//		>>*(ckm.Vus)
//		>>*(ckm.Vub)
//		>>*(ckm.Vcd)
//		>>*(ckm.Vcs)
//		>>*(ckm.Vcb)
//		>>*(ckm.Vtd)
//		>>*(ckm.Vts)
//		>>*(ckm.Vtb)
//		;
////	cout	<<"operator>>:"<<ckm	<<endl;
//	return is;
//}
//
//
//std::ostream& CuTe::NS_LHAFUNC::operator<<( std::ostream& os, const CuTe::NS_LHAFUNC::Charge& C){
//	cout	<< "Charge:" << endl
//			<< "FIXME" << endl
////			<< "\tqf:" << endl << C.qf << endl
////			<< "\t L:" << endl << C.l << endl
////			<< "\t R:" << endl << C.r << endl
////			<< "\tLL:" << endl << C.ll << endl
////			<< "\tRR:" << endl << C.rr << endl
////			<< "\t Q:" << endl << C.qq << endl
//			;
//	return os;
//}
//
//
//
//
//template < >
//Matrix< 2, complex<double> > CuTe::NS_LHAFUNC::Pauli_Matrices::Pauli_i< CuTe::NS_LHAFUNC::Pauli_Matrices::Pauli_1 >(){
//	Matrix<2,complex<double> > M(complex<double>(0.,0.));
//	M[0][1]=complex<double>(1.,0.);
//	M[1][0]=complex<double>(1.,0.);
//	return M;
//}
//
//template < >
//Matrix< 2, complex<double> > CuTe::NS_LHAFUNC::Pauli_Matrices::Pauli_i< CuTe::NS_LHAFUNC::Pauli_Matrices::Pauli_2 >(){
//	Matrix<2,complex<double> > M(complex<double>(0.,0.));
//	M[0][1]=complex<double>(0,-1.);
//	M[1][0]=complex<double>(0,+1.);
//	return M;
//}
//
//template < >
//Matrix< 2, complex<double> > CuTe::NS_LHAFUNC::Pauli_Matrices::Pauli_i< CuTe::NS_LHAFUNC::Pauli_Matrices::Pauli_3 >(){
//	Matrix<2,complex<double> > M(complex<double>(0.,0.));
//	M[0][0]=complex<double>(+1.,0.);
//	M[1][1]=complex<double>(-1.,0.);
//	return M;
//}
//
//template < >
//Matrix< 2, complex<double> > CuTe::NS_LHAFUNC::Pauli_Matrices::Pauli_i< CuTe::NS_LHAFUNC::Pauli_Matrices::Pauli_Plus >(){
//	Matrix<2,complex<double> > M(Pauli_i<Pauli_1>());
//	M += (Pauli_i<Pauli_2>()*complex<double>(0,+1.));
//	M *= 0.5;
//	return M;
//}
//
//template < >
//Matrix< 2, complex<double> > CuTe::NS_LHAFUNC::Pauli_Matrices::Pauli_i< CuTe::NS_LHAFUNC::Pauli_Matrices::Pauli_Minus >(){
//	Matrix<2,complex<double> > M(Pauli_i<Pauli_1>());
//	M -= (Pauli_i<Pauli_2>()*complex<double>(0,+1.));
//	M *= 0.5;
//	return M;
//}
//
//const Matrix< 2, complex<double> > CuTe::NS_LHAFUNC::Charge::pauli_three = CuTe::NS_LHAFUNC::Pauli_Matrices::Pauli_i<CuTe::NS_LHAFUNC::Pauli_Matrices::Pauli_3>();
//const Matrix< 2, complex<double> > CuTe::NS_LHAFUNC::Charge::pauli_plus = CuTe::NS_LHAFUNC::Pauli_Matrices::Pauli_i<CuTe::NS_LHAFUNC::Pauli_Matrices::Pauli_Plus>();



void PDF::_delete_own(){
//		NS_Utils::DOUT dout("PDF::_delete_own");
//
//		dout()	<<"_my_pdf\t= "<<_my_pdf	<<endl;
//		dout()	<<"pdf\t= "<<pdf	<<endl;
//		dout()	<<"_my_alphas\t= "<<_my_alphas	<<endl;
//		dout()	<<"_my_as\t= "<<_my_as	<<endl;
//		dout()	<<"a_s\t= "<<a_s	<<endl;
	my_lhapdf.del();
	delete _my_as;		_my_as	= NULL;
}


void PDF::xfxQ(double x, double mu, std::vector<double>& fxQ) const {
	pdf->xfxQ(x,mu,fxQ);
}


double PDF::operator()(const Parton& i, double x, double mu)const{
	if(pdf!=NULL){
		if(sign>0){
			return pdf->xfxQ(i.lha(),x,mu)/x;
		}else{
			return pdf->xfxQ((-i).lha(),x,mu)/x;
		}
	}else{
		return dummy(x);
	}
}


void PDF::create(const string &PDFname, int member,int sign){
	NS_Utils::DOUT dout("PDF::create");
	_delete_own();
	this->sign	= sign;
	my_lhapdf.set(LHAPDF::mkPDF(PDFname,member),LHAPDF::mkAlphaS(PDFname,member));
	_my_as		= new A_strong;
	_my_as->init(*(my_lhapdf.alphas),my_lhapdf.alphas->orderQCD());

	pdf	= my_lhapdf.pdf;
	a_s	= _my_as;
//		dout()	<< "_my_alphas.nf(12)	= "<<_my_alphas->numFlavorsQ(12)<<endl;
//		dout()	<< "_my_as.nf(12)	= "<<_my_as->nf(12)<<endl;
//		dout()	<< "a_s.nf(12)	= "<<a_s->nf(12)<<endl;
//
//		dout()	<<"_my_pdf\t= "<<_my_pdf	<<endl;
//		dout()	<<"pdf\t= "<<pdf	<<endl;
//		dout()	<<"_my_alphas\t= "<<_my_alphas	<<endl;
//		dout()	<<"_my_as\t= "<<_my_as	<<endl;
//		dout()	<<"a_s\t= "<<a_s	<<endl;
}


void LHA_PDFSET::_clear_(){
//	NS_Utils::DOUT dout("LHA_PDFSET::_clear_");
	if(owns_pdfset){
		delete pdfset;
		pdfset	= NULL;
		owns_pdfset	= false;
	}
	if(N!=0){
		if(owns_pdfs_and_as){
			vector<LHAPDF::PDF*>::iterator it_pdfs(pdfs.begin());
			vector<LHAPDF::PDF*>::iterator end_pdfs(pdfs.end());

			while(it_pdfs!=end_pdfs){
				delete (*(it_pdfs++));	// No = NULL
			}

			vector<LHAPDF::AlphaS*>::iterator as_pdfs(alpha_s.begin());
			vector<LHAPDF::AlphaS*>::iterator end_as(alpha_s.end());

			while(as_pdfs!=end_as){
				delete (*(as_pdfs++));	// No = NULL
			}
			owns_pdfs_and_as = false;
		}

		pdfs.clear();
		alpha_s.clear();

		N	= 0;
	}
}

A_strong& LHA_PDFSET::_init_as(){
	as.resize(N);
	vector<A_strong>::iterator it(as.begin()),end(as.end());
	vector<LHAPDF::AlphaS*>::const_iterator ti(alpha_s.begin());

	while(it!=end){
		(*(it++)).init(*(*ti),(*ti)->orderQCD());
		++ti;
	}

	return as[0];
}

A_strong& LHA_PDFSET::init(const LHAPDF::PDFSet* lha_pdfset, const bool with_errors, int member, const int lhapdf_verbosity){
//	NS_Utils::DOUT dout("LHA_PDFSET::init","string");
	LHAPDF::setVerbosity(lhapdf_verbosity);
	if(lhapdf_verbosity>0){
		cout	<<"PDFsets at "<<LHAPDF::pdfsetsPath()<<endl;
	}
	this->with_errors	= with_errors;
	this->member		= member;

	_clear_();

	pdfset	= lha_pdfset;
	const string PDFname(pdfset->name());


	if(with_errors){
		pdfset->mkPDFs(pdfs);
		N	= pdfs.size();
		alpha_s.reserve(N);

		for(member=0; member<N; ++member){
			alpha_s.push_back(LHAPDF::mkAlphaS(PDFname,member));
		}
	}else{
		pdfs.push_back(pdfset->mkPDF(member));
		alpha_s.push_back(LHAPDF::mkAlphaS(PDFname,member));
		N	= pdfs.size();
		//			dout()	<<"PDF = "<<pdfs[0]<<endl;
	}

	owns_pdfs_and_as	= true;

	return _init_as();
}

A_strong& LHA_PDFSET::init(const string& PDFname, const bool with_errors, const int member, const int lhapdf_verbosity){
	init(new LHAPDF::PDFSet(PDFname),with_errors,member,lhapdf_verbosity);
	owns_pdfset	= true;
	return as[0];
}

A_strong& LHA_PDFSET::init(OWN_LHAPDF *own){
	//	NS_Utils::DOUT dout("LHA_PDFSET::init","name,errors,acc,member");
	this->with_errors	= false;
	this->member		= 0;

	_clear_();

	pdfs.push_back(own->pdf);
	alpha_s.push_back(own->alphas);
	N	= pdfs.size();

	return _init_as();
}


//int LHA_PDFSET::AlphaS_OrderQCD()const{
//	return string_to_T<int>(pdfset->get_entry("AlphaS_OrderQCD",pdfset->get_entry("OrderQCD")));
//}
//
//float LHA_PDFSET::MZ()const{
//	return string_to_T<float>(pdfset->get_entry("MZ"));
//}
//
//float LHA_PDFSET::AlphaS_MZ()const{
//	return string_to_T<float>(pdfset->get_entry("AlphaS_MZ"));
//}
//
//bool LHA_PDFSET::FlavorScheme()const{
//	string tmp(pdfset->get_entry("AlphaS_FlavorScheme",pdfset->get_entry("FlavorScheme")));
//	if(tmp == "fixed") return false;
//	return true;
//}
//
//int LHA_PDFSET::NumFlavors()const{
//	return string_to_T<int>(pdfset->get_entry("AlphaS_NumFlavors",pdfset->get_entry("NumFlavors")));
//}
//
//void LHA_PDFSET::Thresholds(vector<double>& thresholds)const{
//	thresholds.resize(7);
//	thresholds[1]	= string_to_T<double>(pdfset->get_entry("Mup","0."));
//	thresholds[2]	= string_to_T<double>(pdfset->get_entry("MDown","0."));
//	thresholds[3]	= string_to_T<double>(pdfset->get_entry("MStrange","0."));
//	thresholds[4]	= string_to_T<double>(pdfset->get_entry("MCharm"));
//	thresholds[5]	= string_to_T<double>(pdfset->get_entry("MBottom"));
//	thresholds[6]	= string_to_T<double>(pdfset->get_entry("MTop","1e10"));
//}

const PDFUncertainty& LHA_PDFSET::set_uncertainty(const vector<double>& values){
//		NS_Utils::DOUT dout("LHA_PDFSET::set_uncertainty");
	LHAPDF::PDFUncertainty lha;
	pdfset->uncertainty(lha,values);
	errors.central	= lha.central;
	errors.errminus	= lha.errminus;
	errors.errplus	= lha.errplus;
	errors.errsymm	= lha.errsymm;
	errors.scale	= lha.scale;
	return errors;
}


double LHA_PDFSET::get_uncertainty_central()const{
	return errors.central;
}
double LHA_PDFSET::get_uncertainty_errminus()const{
	return errors.errminus;
}
double LHA_PDFSET::get_uncertainty_errplus()const{
	return errors.errplus;
}
double LHA_PDFSET::get_uncertainty_errsymm()const{
	return errors.errsymm;
}
double LHA_PDFSET::get_uncertainty_scale()const{
	return errors.scale;
}
